import os
import subprocess
from collections import OrderedDict
from subprocess import PIPE
from pathlib import Path
import re
from npf import variable, npf
from npf.types.dataset import Run, Dataset
import copy

renametable = {
    'npf.script': 'npf.testie',
}


def mapname(name):
    if name in renametable:
        return renametable[name]
    return name


def mapped_load_global(self):
    module = mapname(self.readline()[:-1])
    name = mapname(self.readline()[:-1])
    klass = self.find_class(module, name)
    self.append(klass)


class Build:
    def __init__(self, repo, version, result_path=None):
        self.n_tests = 0
        self.n_passed = 0
        self.repo = repo
        self.version = version
        self._pretty_name = None
        self._marker = '.'
        self._line = '-'
        self.cache = {}
        self._result_path = result_path

    def copy(self):
        return copy.copy(self)

    def pretty_name(self):
        if self._pretty_name:
            return self._pretty_name
        else:
            return self.version

    @staticmethod
    def __read_file(fp):
        try:
            with open(fp, 'r') as myfile:
                data = myfile.read().replace('\n', '')
        except FileNotFoundError:
            return None
        return data.strip()

    def result_folder(self):
        return (self._result_path[0] if self._result_path else "results") + '/' + self.repo.get_identifier() + '/'

    def result_path(self, test_name, ext, suffix='',prefix=None, folder=''):
        path = self.result_folder() + str(self.version) + ('/' + folder if folder else '') + '/' + (prefix if prefix else '') + os.path.splitext(test_name)[
            0] + suffix + '.' + ext
        return path

    @staticmethod
    def __write_file(fp, val):
        f = open(fp, 'w+')
        f.write(val)
        f.close()

    def __repr__(self):
        return "Build(repo=" + str(self.repo) + ", version=" + self.pretty_name() + ")"

    def __resultFilename(self, script=None):
        if script:
            return self.result_folder() + self.version + '/' + script.filename + ".results"
        else:
            return self.result_folder() + self.version + '.results'

    def writeversion(self, testie, all_results: Dataset, allow_overwrite: bool = False, kind = False, reload=True):
        if not reload and all_results:
          prev=self.load_results(testie = testie, kind = kind, cache=True)
          if prev and len(all_results) < len(prev):
            print("ERROR ! Have less results than before. Forcing update write !")
            reload = True
            return
        if reload:
            results = self.load_results(testie = testie, kind = kind, cache=False)
            if results:
                results.update(all_results)
                all_results = results

        if kind:
            for kind, kresult in all_results.items():
                filename = self.__resultFilename(testie) + '-' + kind
                self._writeversion(filename, kresult, allow_overwrite)
        else:
            filename = self.__resultFilename(testie)
            self._writeversion(filename, all_results, allow_overwrite)

    def _writeversion(self, filename, all_results, allow_overwrite):
        try:
            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
        except OSError:
            print("Error : could not create %s" % os.path.dirname(filename))
        if not allow_overwrite and os.path.exists(filename):
            raise Exception("I refuse to overwrite %s" % filename)
        f = open(filename, 'w+')
        f.seek(0)
        for run, results in all_results.items():
            v = []
            for key, val in sorted(run.variables.items()):
                if type(val) is tuple:
                    val = val[1]
                v.append((key + ":" + str(val).replace(':','\:')).replace(',','\,'))
            type_results = []
            for t,r in results.items():
                str_results = []
                if r is None:
                    pass
                else:
                    for val in r:
                        if type(val) is list:
                            str_results.extend([str(v) for v in val])
                        else:
                            str_results.append(str(val))
                type_results.append(t+':'+(','.join(str_results)))
            f.write(','.join(v) + "={" + '},{'.join(type_results) + "}\n")
        f.close()
        self.cache[filename] = all_results


    def load_results(self, testie, kind=False, cache=True):
        if kind:
            kr={}
            filename = self.__resultFilename(testie) + '-'
            if os.path.exists(os.path.dirname(filename)):
              for f in os.listdir(os.path.dirname(filename)):
                if os.path.basename(filename) in f:
                    kind = f[f.rfind("-") + 1 :]
                    f = filename + kind
                    kr[kind] = self._load_results(testie, f, cache)
            return kr

        else:
            filename = self.__resultFilename(testie)
            return self._load_results(testie, filename, cache)

    def _load_results(self, testie, filename, cache):
        if not Path(filename).exists():
            return None
        if cache:
            if filename in self.cache:
                return self.cache[filename]
        f = open(filename, 'r')
        all_results = OrderedDict()
        try:
            for iline,line in enumerate(f):
                if not line.strip():
                    continue
                variables_data, results_data = line.strip().split('=')

                variables = OrderedDict()

                for v_data in re.split(r'(?<!\\),', variables_data):
                    if v_data.strip():
                        k, v = re.split(r'(?<!\\):', v_data)
                        variables[k] = variable.get_numeric(v) if testie.variables.is_numeric(k) else str(v).replace('\:',':')
                results = {}

                results_data = results_data.strip()[1:-1].split('},{')
                if len(results_data) == 1 and results_data[0].strip() == '':
                    pass
                else:
                    for type_r, results_type_data in [x.split(':') for x in results_data]:
                        results_type_data = results_type_data.split(',')
                        if len(results_type_data) == 1 and results_type_data[0].strip() == '':
                            type_results = None
                        else:
                            type_results = []
                            for result in results_type_data:
                                type_results.append(float(result.strip()))
                        results[type_r] = type_results
                all_results[Run(variables)] = results
        except:
            print("Could not parse %s. The program will stop to avoid erasing data. Please correct or delete the file.\nLine %d : %s" % (filename,iline, line))
            raise
        f.close()
        self.cache[filename] = all_results
        return all_results

    def hasResults(self, script=None):
        return os.path.exists(self.__resultFilename(script))

    def writeResults(self):
        filename = self.__resultFilename()
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        open(filename, 'a').close()

    def build_path(self):
        return self.repo.get_build_path()

    def checkout(self,quiet = False):
        if not self.repo.url:
            return True
        if not self.repo.method.checkout(self.version):
            return False
        self.__write_file(self.build_path() + '/.checkout_version', self.version)
        return True


    def is_checkout_needed(self):
        c_ver = Build.__read_file(self.repo.get_build_path() + '/.checkout_version')
        if c_ver is not None and c_ver == self.version:
            return False
        else:
            return True


    def is_compile_needed(self):
        bin_path=npf.replace_path(self.repo.get_bin_path(self.version),build=self)
        if not os.path.exists(bin_path):
            return "could not find %s" % bin_path
        elif not Build.get_current_version(self.repo) == self.version:
            return "current version is %s, the newest is %s" % (Build.get_current_version(self.repo), self.version)
        else:
            return None

    def compile(self, quiet = False, show_cmd = False):
        """
        Compile the currently checked out repo, assuming it is currently at self.version
        :return: True upon success, False if not
        """
        if not self.repo.url:
            return True
        pwd = os.getcwd()
        os.chdir(self.build_path())

        if self.repo.build_info:
            print(self.repo.build_info)

        for what,command in [("Configuring %s..." % self.version,npf.replace_path(self.repo.configure,build=self)),
                             ("Cleaning %s..." % self.version,npf.replace_path(self.repo.clean,build=self)),
                             ("Building %s..." % self.version,npf.replace_path(self.repo.make,build=self))]:
            if not command:
                continue
            if not quiet:
                print(what)
            if show_cmd and command:
                print(command)
            env = os.environ.copy()
            env.update(self.repo.env)
            p = subprocess.Popen(command, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=env)
            output, err = [x.decode() for x in p.communicate()]
            p.wait()
            if not p.returncode == 0:
                print("Aborted (error code %d) !" % p.returncode)
                print("stdout :")
                print(output)
                print("stderr :")
                print(err)
                self.__write_file('.build_version', '')
                os.chdir(pwd)
                return False

        os.chdir(pwd)
        self.__write_file(Build.__get_build_version_path(self.repo), self.version)
        return True

    def build(self, force_build : bool = False, never_build : bool = False, quiet_build : bool = False, show_build_cmd : bool = False, executor=None):
        if force_build or self.is_checkout_needed():
            force_build = "there is a new version available"
            if never_build:
                if not quiet_build:
                    print("Warning : test will be done with an unknown state of build")
            if not quiet_build:
                print("Checking out %s" % (self.repo.name))
            if not self.checkout(quiet_build):
                return False
        reason = self.is_compile_needed()
        if force_build or (reason != None):
            if never_build:
                print("Warning : version changed but you disallowed build. Test will be done with an unknown state of build")
                self.repo._current_build = self
                return True
            if not quiet_build:
                print("Building %s (because %s)" % (self.repo.name, "you force the build with --force-build" if force_build is True else reason if not force_build else force_build ))
            if not self.compile(quiet_build, show_build_cmd):
                return False
        self.repo._current_build = self
        return True

    def get_bin_folder(self):
        return self.repo.get_bin_folder(self.version)

    def __str__(self):
        return "Build(repo = %s, version = %s)" % (self.repo,self.version)

    @staticmethod
    def get_current_version(repo):
        return Build.__read_file(Build.__get_build_version_path(repo))

    @staticmethod
    def __get_build_version_path(repo):
        return repo.get_build_path() + '/.build_version'



import os
import pwd
import signal
from multiprocessing import Queue, Event
from subprocess import PIPE, Popen, TimeoutExpired
from typing import List
from .executor import Executor

class LocalKiller:
    def __init__(self, pgpid):
        self.pgpid = pgpid

    def kill(self):
        os.killpg(self.pgpid, signal.SIGKILL)

    def force_kill(self):
        os.killpg(self.pgpid, signal.SIGTERM)

    def is_alive(self):
        try:
            os.killpg(self.pgpid, 0)
        except PermissionError:
            return True
        except ProcessLookupError:
            return False
        return True

class LocalExecutor(Executor):
    def __init__(self):
        super().__init__()

    def exec(self, cmd, bin_paths : List[str]=[], queue: Queue = None, options = None, stdin = None, timeout = None, sudo = False, testdir=None, event=None, title=None, env = {}, virt="" ):
        if testdir is not None:
            os.chdir("..")
        if not title:
            title = "local"
        cwd = os.getcwd()
        env = env.copy()
        env.update(os.environ)
        if bin_paths:
            if not sudo:
                env["PATH"] = ':'.join([cwd + '/' + path if not os.path.abspath(path) else path for path in bin_paths]) + ":" + env["PATH"]
            else:
                cmd = 'export PATH=' + ':'.join([cwd + '/' + path if not os.path.abspath(path) else path for path in bin_paths]) + ":" + '$PATH\n' + cmd

        if options is not None and options.show_cmd:
            print("Executing (PATH+=%s) :\n%s" % (':'.join(bin_paths), cmd.strip()))

        if sudo and pwd.getpwuid(os.getuid()).pw_name != "root":
            cmd = "sudo -E " + virt + "  bash -c '"+ cmd.replace("'", "'\"'\"'") + "'";
        else:
            cmd = virt + " bash -c '"+ cmd.replace("'", "'\"'\"'") + "'";


        p = Popen(cmd,
                  stdin=PIPE, stdout=PIPE, stderr=PIPE,
                  shell=True, preexec_fn=os.setsid,
                  env=env)
        pid = p.pid
        pgpid = os.getpgid(pid)

        killer = LocalKiller(pgpid)
        if queue:
            queue.put(killer)
        try:
            s_output, s_err = [x.decode() for x in
                               p.communicate(input = stdin,  timeout=timeout)]
            self.searchEvent(s_output, event)
            if options and options.show_full:
                for line in s_output.splitlines():
                    self._print(title, line, True)
            p.stdin.close()
            p.stderr.close()
            p.stdout.close()
            if testdir is not None:
                os.chdir(testdir)
            return pid, s_output, s_err, 0 if event and event.is_terminated() else p.returncode
        except TimeoutExpired:
            print("Test expired")
            p.terminate()
            p.kill()
            os.killpg(pgpid, signal.SIGKILL)
            os.killpg(pgpid, signal.SIGTERM)
            s_output, s_err = [x.decode() for x in p.communicate()]
            p.stdin.close()
            p.stderr.close()
            p.stdout.close()
            if testdir is not None:
                os.chdir(testdir)
            return 0, s_output, s_err, p.returncode
        except KeyboardInterrupt:
            os.killpg(pgpid, signal.SIGKILL)
            if testdir is not None:
                os.chdir(testdir)
            return -1, s_output, s_err, p.returncode

    def writeFile(self,filename,path_to_root,content):
        f = open(filename, "w")
        f.write(content)
        f.close()
        return True



#! /usr/bin/env python3
"""
Main NPF testie runner program
"""
import argparse
import errno

import sys

from npf import npf
from npf.regression import *
from npf.statistics import Statistics
from npf.testie import Testie, ScriptInitException


def main():
    parser = argparse.ArgumentParser(description='NPF Testie runner')
    v = npf.add_verbosity_options(parser)

    b = npf.add_building_options(parser)
    b.add_argument('--allow-old-build',
                   help='Re-build and run test for old versions (compare-version and graph-version) without results. '
                        'By default, only building for the regression versions (see --history or --version) is done',
                   dest='allow_oldbuild', action='store_true', default=False)
    b.add_argument('--force-old-build',
                   help='Force to rebuild the old versions. Ignored if allow-old-build is not set', dest='force_oldbuild',
                   action='store_true', default=False)

    t = npf.add_testing_options(parser, True)

    g = parser.add_argument_group('Versioning options')
    g.add_argument('--regress',
                    help='Do a regression comparison against old version of the software', dest='compare', action='store_true',
                    default=False)
    gf = g.add_mutually_exclusive_group()
    gf.add_argument('--history',
                    help='Number of commits in the history on which to execute the regression tests. By default, '
                         'this is 1 meaning that the regression test is done on HEAD, and will be compared '
                         'against HEAD~1. This parameter allows to '
                         'start at commits HEAD~N as if it was HEAD, doing the regression test for each'
                         'commits up to now. Difference with --allow-old-build is that the regression test '
                         'will be done for each commit instead of just graphing the results, so error message and'
                         'return code will concern any regression between HEAD and HEAD~N. '
                         'Ignored if --version is given.',
                    dest='history', metavar='N',
                    nargs='?', type=int, default=1)
    g.add_argument('--branch', help='Branch', type=str, nargs='?', default=None)
    g.add_argument('--compare-version', dest='compare_version', metavar='version', type=str, nargs='?',
                   help='A version to compare against the last version. Default is the first parent of the last version containing some results.')


    a = npf.add_graph_options(parser)
    af = a.add_mutually_exclusive_group()
    af.add_argument('--graph-version', metavar='version', type=str, nargs='*',
                    help='versions to simply graph')
    af.add_argument('--graph-num', metavar='N', type=int, nargs='?', default=-1,
                    help='Number of olds versions to graph after --compare-version, unused if --graph-version is given. Default is 0 or 8 if --regress is given.')
    # a.add_argument('--graph-allvariables', help='Graph only the latest variables (usefull when you restrict variables '
    #                                             'with tags)', dest='graph_newonly', action='store_true', default=False)
    # a.add_argument('--graph-serie', dest='graph_serie', metavar='variable', type=str, nargs=1, default=[None],
    #                 help='Set which variable will be used as serie when creating graph');

    parser.add_argument('repo', metavar='repo name', type=str, nargs='?', help='name of the repo/group of builds', default=None)

    args = parser.parse_args()


    npf.parse_nodes(args)


    if args.force_oldbuild and not args.allow_oldbuild:
        print("--force-old-build needs --allow-old-build")
        parser.print_help()
        return 1

    if args.repo:
        repo = Repository.get_instance(args.repo, args)
    else:
        if os.path.exists(args.testie) and os.path.isfile(args.testie):
            tmptestie = Testie(args.testie,options=args)
            if "default_repo" in tmptestie.config and tmptestie.config["default_repo"] is not None:
                repo = Repository.get_instance(tmptestie.config["default_repo"], args)
            else:
                print("This testie has no default repository")
                sys.exit(1)
        else:
            print("Please specify a repository to use to the command line or only a single testie with a default_repo")
            sys.exit(1)

    if args.graph_num == -1:
        args.graph_num = 8 if args.compare else 0


    tags = args.tags
    tags += repo.tags

    #Overwrite config if a build folder is given
    if args.use_local:
        repo.url = None
        repo._build_path = args.use_local + '/'
        versions = ['local']
    elif repo.url:
        versions = repo.method.get_last_versions(limit=args.history,branch=args.branch)
    else:
        versions = ['local']

    # Builds of the regression versions
    builds = []

    for version in versions:
        builds.append(Build(repo, version))

    last_rebuilds = []

    last_build = None
    if args.compare:
        if args.compare_version and len(args.compare_version):
            compare_version = args.compare_version
            last_build = Build(repo, compare_version)
        else:
            old_versions = repo.method.get_history(versions[-1],100)
            for i, version in enumerate(old_versions):
                last_build = Build(repo, version)
                if last_build.hasResults():
                    break
                elif args.allow_oldbuild:
                    last_rebuilds.append(last_build)
                    break
                if i > 100:
                    last_build = None
                    break
            if last_build:
                print("Comparaison version is %s" % last_build.version)

    graph_builds = []
    if args.graph_version and len(args.graph_version) > 0:
        for g in args.graph_version:
            graph_builds.append(Build(repo, g))
    else:
        if args.graph_num > 1 and repo.url:
            old_versions = repo.method.get_history(last_build.version if last_build else builds[-1].version, 100)
            for i, version in enumerate(old_versions):
                g_build = Build(repo, version)
                if g_build in builds or g_build == last_build:
                    continue
                i += 1
                if g_build.hasResults() and not args.force_oldbuild:
                    graph_builds.append(g_build)
                elif args.allow_oldbuild:
                    last_rebuilds.append(g_build)
                    graph_builds.append(g_build)
                if len(graph_builds) > args.graph_num:
                    break

    testies = Testie.expand_folder(testie_path=args.testie, options=args, tags=tags)
    if not testies:
        sys.exit(errno.ENOENT)

    npf.override(args, testies)

    for b in last_rebuilds:
        print("Last version %s had no result. Re-executing tests for it." % b.version)
        did_something = False
        for testie in testies:
            prev_results = b.load_results(testie)
            print("Executing testie %s" % testie.filename)
            try:
                all_results, time_results, init_done = testie.execute_all(b,options=args, prev_results=prev_results)

                if all_results is None and time_results is None:
                    continue
            except ScriptInitException:
                continue
            else:
                did_something = True
            b.writeversion(testie, all_results, allow_overwrite=True)
        if did_something:
            b.writeResults()

    returncode = 0

    for build in reversed(builds):
        if len(builds) > 1 or repo.version:
            print("Starting tests for version %s" % build.version)

        nok = 0
        ntests = 0

        for testie in testies:
            print("Executing testie %s" % testie.filename)

            regression = Regression(testie)

            print(testie.get_title())

            old_all_results = None
            if last_build:
                try:
                    old_all_results = last_build.load_results(testie)
                except FileNotFoundError:
                    print("Previous build %s could not be found, we will not compare !" % last_build.version)
                    last_build = None

            try:
                prev_results = build.load_results(testie)
                prev_kind_results = build.load_results(testie, kind=True)
            except FileNotFoundError:
                prev_results = None
                prev_kind_results = None

            all_results = None
            time_results = None
            try:
                if all_results is None and time_results is None:
                    all_results, time_results, init_done = testie.execute_all(build, prev_results=prev_results, prev_kind_results=prev_kind_results, do_test=args.do_test, options=args)
                if not all_results and not time_results:
                    returncode+=1
                    continue
            except ScriptInitException:
                continue

            if args.compare:
                variables_passed,variables_passed = regression.compare(testie, testie.variables, all_results, build, old_all_results, last_build)
                if variables_passed == variables_passed:
                    nok += 1
                else:
                    returncode += 1
                ntests += 1

            if all_results and len(all_results) > 0:
                build.writeResults()

            #Filtered results are results only for the given current variables
            filtered_results = {}
            for v in testie.variables:
                run = Run(v)
                if run in all_results:
                    filtered_results[run] = all_results[run]

            if args.statistics:
                Statistics.run(build,filtered_results, testie, max_depth=args.statistics_maxdepth, filename=args.statistics_filename)

            grapher = Grapher()

            g_series = []
            if last_build and old_all_results and args.compare:
                g_series.append((testie, last_build, old_all_results))

            for g_build in graph_builds:
                try:
                    g_all_results = g_build.load_results(testie)
                    if (g_all_results and len(g_all_results) > 0):
                        g_series.append((testie, g_build, g_all_results))
                except FileNotFoundError:
                    print("Previous build %s could not be found, we will not graph it !" % g_build.version)

            filename = args.graph_filename if args.graph_filename else build.result_path(testie.filename, 'pdf')
            grapher.graph(series=[(testie, build, all_results)] + g_series,
                          title=testie.get_title(),
                          filename=filename,
                          graph_variables=[Run(x) for x in testie.variables],
                          options = args)
            if time_results:
                for find, results in time_results.items():
                    if not results:
                        continue
                    grapher.graph(series=[(testie, build, results)],
                          title=testie.get_title(),
                          filename=filename,
                          options = args)
        if last_build and args.graph_num > 0:
            graph_builds = [last_build] + graph_builds[:-1]
        last_build = build
        if args.compare:
            print("[%s] Finished run for %s, %d/%d tests passed" % (repo.name, build.version, nok, ntests))

    sys.exit(returncode)


if __name__ == "__main__":
    main()



import os
import tarfile
import urllib
from collections import OrderedDict
from abc import ABCMeta
from pathlib import Path
import re
from urllib.error import URLError

import shutil

import gitdb

from npf import npf
from npf.build import Build
from .variable import is_numeric, VariableFactory

import git

repo_variables = ['name', 'branch', 'configure', 'url', 'method', 'parent', 'tags', 'make', 'version', 'clean', 'build_info',
                  'bin_folder', 'bin_name', 'env']


class Method(metaclass=ABCMeta):
    def __init__(self, repo):
        self.repo = repo


class MethodGit(Method):
    __gitrepo = None
    _fetch_done = False

    def gitrepo(self) -> git.Repo:
        if (self.__gitrepo):
            return self.__gitrepo
        else:
            return self.checkout()

    def get_last_versions(self, limit=100, branch=None, force_fetch=False):
        versions = []
        origin = self.gitrepo().remote('origin')
        if not self.repo.options.no_build and (force_fetch or not self._fetch_done):
            if not self.repo.options.quiet_build:
                print("Fetching last versions of %s..." % self.repo.reponame)
            origin.fetch()
            self._fetch_done = True

        if branch is None:
            branch = self.repo.branch

        try:
            b_commit = self.gitrepo().commit('origin/' + branch)
        except gitdb.exc.BadName:
            b_commit = self.gitrepo().tag('refs/tags/' + branch).commit

        for i, commit in enumerate(b_commit.iter_items(repo=b_commit.repo, rev=b_commit, skip=0)):
            versions.append(commit.hexsha[:7])
            if (len(versions) >= limit):
                break
        return versions

    def get_history(self, version, limit=1):
        versions = []
        for commit in next(self.gitrepo().iter_commits(version)).iter_parents():
            versions.append(commit.hexsha[:7])
            if len(versions) == limit:
                break
        return versions

    def is_checkout_needed(self, version):
        gitrepo = self.gitrepo()
        if gitrepo.head.commit.hexsha[:7] != version:
            return True
        return False

    def checkout(self, branch=None) -> git.Repo:
        """
        Checkout the repo to its folder, fetch if it already exists
        :param branch: An optional branch
        :return:
        """
        if not branch:
            branch = self.repo.branch

        need_clone=False
        if os.path.exists(self.repo.get_build_path()):
            try:
                gitrepo = git.Repo(self.repo.get_build_path())
                o = gitrepo.remotes.origin
                if not self.repo.options.no_build and not self._fetch_done:
                    o.fetch()
                    self._fetch_done = True
            except git.exc.InvalidGitRepositoryError:
                print("Path %s appear to be invalid" % self.repo.get_build_path())
                #shutil.rmtree(self.repo.get_build_path())
                need_clone=True
        else:
            need_clone=True

        if need_clone:
            if not self.repo.options.quiet_build:
                print("Cloning %s from %s..." % (self.repo.reponame, self.repo.url))
            gitrepo = git.Repo.clone_from(self.repo.url, self.repo.get_build_path())

        if branch in gitrepo.remotes.origin.refs:
            c = gitrepo.remotes.origin.refs[branch]
        else:
            c = branch

        if gitrepo.head.commit != gitrepo.commit(c) and not self.repo.options.no_build:
            if not self.repo.options.quiet_build:
                print("Reseting branch to latest %s" % (c))
            gitrepo.git.stash('save')
            gitrepo.head.reset(commit=c, index=True, working_tree=True)

        self.__gitrepo = gitrepo
        return gitrepo


class UnversionedMethod(Method, metaclass=ABCMeta):
    def __init__(self, repo):
        super().__init__(repo)
        if not repo.version:
            raise Exception("This method require a version")

    def get_last_versions(self, limit=None, branch=None, force_fetch=False):
        return [self.repo.version]

    def get_history(self, version, limit):
        return []


class MethodGet(UnversionedMethod):
    def checkout(self, branch=None):
        if branch is None:
            branch = self.repo.version
        url = npf.replace_path(self.repo.url,Build(self.repo,branch,self.options.result_path))
        if not Path(self.repo.get_build_path()).exists():
            os.makedirs(self.repo.get_build_path())
        try:
            filename, headers = urllib.request.urlretrieve(url, self.repo.get_build_path() + os.path.basename(url))
        except URLError:
            print("ERROR : Could not download %s : bad URL?" % url)
            return False
        t = tarfile.open(filename)
        t.extractall(self.repo.get_build_path())
        t.close()
        os.unlink(filename)
        return True


class MethodPackage(UnversionedMethod):
    def checkout(self, branch=None):
        if not Path(self.repo.get_build_path()).exists():
            os.makedirs(self.repo.get_build_path())
        return True


repo_methods = {'git': MethodGit, 'get': MethodGet, 'package': MethodPackage}


class Repository:
    _repo_cache = {}


    method = repo_methods['git']

    def __init__(self, repo, options):
        self.name = None
        self._current_build = None

        version = repo.split('@')
        if len(version) > 1:
            self.version=version[1]
        else:
            self.version=None
        overwrite_name = version[0].split(':')
        add_tags = overwrite_name[0].split('+')
        overwrite_branch = add_tags[0].split('/')

        self.reponame = overwrite_branch[0]
        self.tags = []
        self.options = options
        self.branch = 'master'
        self.make = 'make -j12'
        self.clean = 'make clean'
        self.bin_folder = 'bin'
        self.env = OrderedDict()
        self.bin_name = self.reponame  # Wild guess that may work some times...
        self.build_info = None
        self.configure = ''
        self._last_100 = None

        if self.reponame == 'None':
            self.url = None
        else:
            repo_path = npf.find_local('repo/' + self.reponame + '.repo')

            f = open(repo_path, 'r')
            for line in f:
                line = line.strip()
                line = re.sub(r'(^|[ ])//.*$', '', line)
                if line.startswith("#"):
                    continue
                if not line:
                    continue
                s = line.split('=', 1)
                var = s[0].strip()
                var = var.split(':',1)
                if len(var) > 1:
                    have_all=True
                    for v in var[0].split(','):
                        if not v in options.tags:
                            have_all = False
                            break
                    if not have_all:
                        continue
                    self.reponame += '-' + var[0]
                    var = var[1]
                else:
                    var = var[0]
                val = s[1].strip()
                append = False
                if var.endswith('+'):
                    var = var[:-1]
                    append = True

                if is_numeric(val) and var != 'branch':
                    val = float(val)
                    if val.is_integer():
                        val = int(val)
                if not var in repo_variables:
                    raise Exception("Unknown variable %s" % var)
                elif var == "parent":
                    parent = Repository(val, options)
                    for attr in repo_variables:
                        if not hasattr(parent, attr):
                            continue
                        pval = getattr(parent, attr)
                        if attr == "method":
                            for m, c in repo_methods.items():
                                if c == type(pval):
                                    method = m
                                    break
                        else:
                            setattr(self, attr, pval)
                elif var == "method":
                    val = val.lower()
                    if not val in repo_methods:
                        raise Exception("Unknown method %s" % val)
                    val = repo_methods[val]
                elif var == "tags":
                    if append:
                        self.tags += val.split(',')
                    else:
                        self.tags = val.split(',')
                    continue

                elif var == "env":
                    ed = VariableFactory.build(var, val).vdict
                    if append:
                        self.env += ed
                    else:
                        self.env = ed
                    continue

                if append:
                    setattr(self, var, getattr(self, var) + " " + val)
                else:
                    setattr(self, var, val)

            self.method = self.method(self)  # Instanciate the method

        self.overriden_variables = {}
        if len(add_tags) > 1:
            for spec in add_tags[1].split(','):
                sp = spec.split('=')
                if len(sp) == 1:
                    self.tags.append(spec)
                else:
                    self.overriden_variables[sp[0]] = sp[1]
            self._id = self.reponame + "-" + add_tags[1]
        else:
            self._id = self.reponame


        if len(overwrite_branch) > 1:
            self.branch = overwrite_branch[1]

        if len(overwrite_name) > 1:
            self.name = overwrite_name[1]


        self._build_path = os.path.dirname(
            (options.build_folder if not options.build_folder is None else npf.npf_root()+'/build/') + self.reponame + '/')

    def get_identifier(self):
        return self._id

    def get_reponame(self):
        return self.reponame

    def pretty_name(self):
        return self.name

    def get_build_path(self):
        return self._build_path

    def get_bin_folder(self, version=None):
        if version is None:
            version = self.current_version()
        if version is None:
            bin_folder = self.bin_folder
        else:
            bin_folder = self.bin_folder.replace('$version', version)
        return self.get_build_path() + '/' + (bin_folder + '/' if bin_folder else '')

    def get_bin_path(self, version):
        if version is None:
            version = self.current_version()
        if version is None:
            bin_name = self.bin_name
        else:
            bin_name = self.bin_name.replace('$version', version)
        return self.get_bin_folder(version) + bin_name

    def get_last_build(self, history: int = 1, stop_at: Build = None, with_results=False, force_fetch=False) -> Build:
        if self.version:
            versions = [self.version]
        else:
            if force_fetch or not self._last_100:
                versions = self.method.get_last_versions(100, force_fetch=force_fetch)
                self._last_100 = versions
            else:
                versions = self._last_100

        last_build = None
        for i, version in enumerate(versions):
            if stop_at and version == stop_at.version:
                if i == 0:
                    return None
                break
            last_build = Build(self, version, self.options.result_path)
            if not with_results or last_build.hasResults():
                if history <= 1:
                    break
                else:
                    history -= 1
            if i > 100:
                last_build = None
                break
        return last_build

    def last_build_before(self, old_build) -> Build:
        return self.get_last_build(stop_at=old_build, history=-1)

    def get_old_results(self, last_graph: Build, num_old: int, testie):
        graphs_series = []
        # Todo
        parents = self.method.gitrepo().iter_commits(last_graph.version)
        next(parents)  # The first commit is last_graph itself

        for i, commit in enumerate(parents):  # Get old results for graph
            g_build = Build(self, commit.hexsha[:7], self.options.result_path)
            if not g_build.hasResults(testie):
                continue
            g_all_results = g_build.load_results(testie)
            graphs_series.append((testie, g_build, g_all_results))
            if (i > 100 or len(graphs_series) == num_old):
                break
        return graphs_series

    def current_build(self):
        if self._current_build:
            return self._current_build
        return None

    def current_version(self):
        build = self.current_build()
        if build:
            return build.version
        return Build.get_current_version(self)

    def __str__(self):
        return self.get_reponame()

    @classmethod
    def get_instance(cls, dep, options):
        if dep in cls._repo_cache:
            return cls._repo_cache[dep]
        try:
            repo = Repository(dep, options)
        except FileNotFoundError:
            raise Exception("%s is not a valid repository name !" % (dep))
        cls._repo_cache[dep] = repo
        return repo



import multiprocessing
import os
import sys
import threading
import time
import traceback
import random
import shutil
import datetime
import itertools
import string
from pathlib import Path
from queue import Empty
from typing import Tuple, Dict
import numpy as np
from npf.build import Build
from npf.node import NIC
from npf.section import *
from npf.npf import get_valid_filename
from npf.types.dataset import Run, Dataset
from npf.eventbus import EventBus
from .variable import get_bool
from decimal import *

from subprocess import PIPE, Popen, TimeoutExpired

class RemoteParameters:
    def __init__(self):
        self.default_role_map = None
        self.role = None
        self.delay = None
        self.executor = None
        self.bin_paths = None
        self.queue = None
        self.sudo = None
        self.autokill = None
        self.queue = None
        self.timeout = None
        self.options = None
        self.stdin = None
        self.commands = None
        self.testdir = None
        self.waitfor = None
        self.event = None
        self.title = None
        self.env = None
        self.virt = ""

    pass


def _parallel_exec(param: RemoteParameters):
    nodes = npf.nodes_for_role(param.role)
    executor = nodes[param.role_id].executor
    if param.waitfor:
        n=1
        #print("Waiting for %s" % param.waitfor)
        if param.waitfor[0].isdigit():
            n=int(param.waitfor[0])
            param.waitfor=param.waitfor[1:]
        for i in range(n):
            param.event.listen(param.waitfor)

    param.event.wait_for_termination(param.delay)
    if param.event.is_terminated():
        return 1, 'Killed before execution', 'Killed before execution', 0, param.script
    pid, o, e, c = executor.exec(cmd=param.commands,
                                 stdin=param.stdin,
                                 timeout=param.timeout,
                                 bin_paths=param.bin_paths,
                                 queue=param.queue,
                                 options=param.options,
                                 sudo=param.sudo,
                                 testdir=param.testdir,
                                 event=param.event,
                                 title=param.name,
                                 env=param.env,
                                 virt=param.virt)
    if pid == 0:
        return False, o, e, c, param.script
    else:
        if param.autokill is not None:
                param.event.c.acquire()
                param.autokill.value = param.autokill.value - 1
                v = param.autokill.value
                param.event.c.release()
                if v == 0:
                    Testie.killall(param.queue, param.event)

        elif pid == -1:
            Testie.killall(param.queue, param.event)
        if pid == -1:
            return -1, o, e, c, param.script
        return True, o, e, c, param.script


class ScriptInitException(Exception):
    pass


class Testie:
    def get_name(self):
        return self.filename

    def get_scripts(self) -> List[SectionScript]:
        return self.scripts

    def __init__(self, testie_path, options, tags=None, role=None):
        if os.path.exists(npf.find_local(testie_path)):
            testie_path = npf.find_local(testie_path)
        else:
            if not os.path.exists(npf.find_local(testie_path + '.testie')):
                raise Exception("Could not find testie %s" % testie_path)
            testie_path = npf.find_local(testie_path + '.testie')

        self.sections = []
        self.files = []
        self.init_files = []
        self.late_variables = []
        self.scripts = []
        self.imports = []
        self.requirements = []
        self.sendfile = {}
        self.filename = os.path.basename(testie_path)
        self.path = os.path.dirname(testie_path)
        self.options = options
        self.tags = tags if tags else []
        self.role = role

        i = -1
        try:
            section = None
            f = open(testie_path, 'r')
            for i, line in enumerate(f):
                if section is None or section.noparse is False:
                    line = re.sub(r'(^|[ ])//.*$', '', line)
                if line.startswith('#') and section is None:
                    print("Warning : comments now use // instead of #. This will be soon deprecated")
                    continue
                if line.strip() == '' and not section:
                    continue

                if line.startswith("%"):
                    result = line[1:]
                    section = SectionFactory.build(self, result.strip())

                    if not section is SectionNull:
                        self.sections.append(section)
                elif section is None:
                    raise Exception("Bad syntax, file must start by a section. Line %d :\n%s" % (i, line))
                else:
                    section.content += line
            f.close()

            if not hasattr(self, "info"):
                self.info = Section("info")
                self.info.content = self.filename
                self.sections.append(self.info)

            if not hasattr(self, "stdin"):
                self.stdin = Section("stdin")
                self.sections.append(self.stdin)

            if not hasattr(self, "variables"):
                self.variables = SectionVariable()
                self.sections.append(self.variables)

            if not hasattr(self, "config"):
                self.config = SectionConfig()
                self.sections.append(self.config)

            for section in self.sections:
                section.finish(self)
        except Exception as e:
            if i == -1:
                raise Exception("An exception occured while accessing the file %s" % (testie_path))
            else:
                raise Exception("An exception occured while parsing %s at line %d:\n%s" % (testie_path, i, e.__str__()))

        # Check that all reference roles are defined
        known_roles = {'self', 'default'}.union(set(npf.roles.keys()))
        for script in self.get_scripts():
            known_roles.add(script.get_role())
        # for file in self.files:
        #            for nicref in re.finditer(Node.NICREF_REGEX, file.content, re.IGNORECASE):
        #                if nicref.group('role') not in known_roles:
        #                    raise Exception("Unknown role %s" % nicref.group('role'))

        # Create imports testies
        for imp in self.imports:
            from npf.module import Module
            imp.testie = Module(imp.module, options, self, imp, tags, imp.get_role())
            if len(imp.testie.variables.dynamics()) > 0:
                raise Exception("Imports cannot have dynamic variables. Their parents decides what's dynamic.")
            if 'as_init' in imp.params:
                for script in imp.testie.scripts:
                    script.init = True
            if 'delay' in imp.params:
                delay = imp.params.setdefault('delay', 0)
                for script in imp.testie.scripts:
                    if not 'delay' in script.params:
                        script.params['delay'] = float(delay) + float(imp.params['delay'])
                del imp.params['delay']
            if 'waitfor' in imp.params:
                for script in imp.testie.scripts:
                    if script.type == SectionScript.TYPE_SCRIPT:
                        if not 'waitfor' in script.params:
                            script.params['waitfor'] = imp.params['waitfor']
                del imp.params['waitfor']
            if 'autokill' in imp.params:
                for script in imp.testie.scripts:
                    if script.type == SectionScript.TYPE_SCRIPT:
                        script.params['autokill'] = imp.params['autokill']
                del imp.params['autokill']
            if imp.multi is not None:
                for script in imp.testie.scripts:
                    if script.type == SectionScript.TYPE_SCRIPT or script.type == SectionScript.TYPE_INIT:
                        script.multi = imp.multi
            overriden_variables = {}
            for k, v in imp.params.items():
                overriden_variables[k] = VariableFactory.build(k, v)
            imp.testie.variables.override_all(overriden_variables)
            if not imp.is_include:
                for script in imp.testie.scripts:
                    if script.get_role():
                        raise Exception('Modules cannot have roles, their importer defines it')
                    script._role = imp.get_role()
                nsendfile = {}
                for role,fpaths in imp.testie.sendfile.items():
                    if role and role != 'default':
                        raise Exception('Modules cannot have roles (sendfile has role %s), their importer defines it' % role)
                    nsendfile[imp.get_role()] = fpaths
                imp.testie.sendfile = nsendfile

                if hasattr(imp.testie, 'exit'):
                    imp.testie.exit._role = imp.get_role()
            else: #is include
                self.variables.vlist.update(imp.testie.variables.vlist)

    def build_deps(self, repo_under_test: List[Repository], v_internals={}, no_build=False, done=None):
        if done is None:
            done = set()
        # Check for dependencies
        deps = set()
        for script in self.get_scripts():
            deps = deps.union(script.get_deps())
        for dep in deps:
            if dep in [repo.reponame for repo in repo_under_test]:
                continue
            deprepo = Repository.get_instance(dep, self.options)
            if deprepo.url is None or deprepo.reponame in self.options.no_build_deps:
                continue
            if not deprepo.get_last_build().build(force_build=deprepo.reponame in self.options.force_build_deps):
                raise Exception("Could not build dependency %s" % dep)
        #Do the same for the imports
        for imp in self.imports:
            imp.testie.build_deps(repo_under_test, v_internals, True, done=done)

        # Send dependencies for nfs=0 nodes
        toSend = set()
        for script in self.get_scripts():
            role = script.get_role()
            nodes = npf.nodes_for_role(role)
            for node in nodes:
              if not node.nfs:
                for repo in repo_under_test:
                    if repo.get_reponame() != 'local' and not no_build:
                        toSend.add((repo.reponame,role,node, repo.get_build_path()))
                for dep in script.get_deps():
                    deprepo = Repository.get_instance(dep, self.options)

                    toSend.add((deprepo.reponame,role,node,deprepo.get_build_path()))
        for repo,role,node,bp in toSend.difference(done):
            print("Sending software %s to %s... " % (repo, role), end ='')
            t = node.executor.sendFolder(os.path.relpath(bp,npf.npf_root()),local=npf.npf_root())
            if (t > 0):
                print("%d bytes sent." % t)
            else:
                print("Already up to date !")

        done.update(toSend)

        st = dict([(f, v.makeValues()[0]) for f, v in self.variables.statics().items()])
        st.update(v_internals)

        for late_variables in self.get_late_variables():
            st.update(late_variables.execute(st, self, fail=False))

        L = [imp.testie.sendfile for imp in self.imports]
        for role, fpaths in itertools.chain(self.sendfile.items(), {k: v for d in L for k, v in d.items()}.items()):
            nodes = npf.nodes_for_role(role)
            for node in nodes:
              if not node.nfs:
                for fpath in fpaths:
                    fpath = SectionVariable.replace_variables(st, fpath, role, self_node=node,
                                                              default_role_map=self.config.get_dict("default_role_map"))
                    if not os.path.isabs(fpath):
                        fpath = './npf/' + fpath
                    fpath = os.path.relpath(fpath)
                    print("Sending files %s to %s... " % (fpath, role), end = '')
                    t = node.executor.sendFolder(os.path.relpath(fpath,npf.npf_root()))
                    if (t > 0):
                        print("%d bytes sent." % t)
                    else:
                        print("Already up to date !")


        return True

    def test_tags(self):
        missings = []
        # print("%s requires " % self.get_name(), self.config.get_list("require_tags"))
        for tag in self.config.get_list("require_tags"):
            if not tag in self.tags:
                missings.append(tag)
        return missings

    def build_file_list(self, v, self_role=None, files=None) -> List[Tuple[str, str, str]]:
        list = []
        if files is None:
            files = self.files
        for s in files:
            role = s.get_role() if s.get_role() else self_role
            v["NPF_NODE_MAX"] = len(npf.nodes_for_role(role))
            if not s.noparse:
                p = SectionVariable.replace_variables(v, s.content, role,default_role_map = self.config.get_dict("default_role_map"))
            else:
                p = s.content
            list.append((s.filename, p, role))
        return list

    def create_files(self, file_list, path_to_root):
        unique_list = {}
        for filename, p, role in file_list:
            if filename in unique_list:
                if unique_list[filename + (role if role else '')][1] != p:
                    raise Exception(
                        "File name conflict ! Some of your scripts try to create some file with the same name but "
                        "different content (%s) !" % filename)
            else:
                unique_list[filename + (role if role else '')] = (filename, p, role)

        for whatever, (filename, p, role) in unique_list.items():
            if self.options.show_files:
                print("File %s:" % filename)
                print(p.strip())
            for node in npf.nodes_for_role(role):
                if not node.executor.writeFile(filename, path_to_root, p):
                    raise Exception("Could not create file %s on %s" % (filename, node.name))

    def test_require(self, v, build):
        for require in self.requirements + list(itertools.chain.from_iterable([imp.testie.requirements for imp in self.imports])):
            p = SectionVariable.replace_variables(v, require.content, require.role(),
                                                  self.config.get_dict("default_role_map"))
            pid, output, err, returncode = npf.executor(require.role(), self.config.get_dict("default_role_map")).exec(
                cmd=p, bin_paths=[build.get_bin_folder()], options=self.options, event=None, testdir=None)
            if returncode != 0:
                return False, output, err
            continue
        return True, '', ''

    def cleanup(self):
        for s in self.files:
            path = Path(s.filename)
            if path.is_file():
                path.unlink()

    @staticmethod
    def killall(queue, event):
        event.terminate()
        while not queue.empty():
            try:
                killer = queue.get(block=False)
            except Empty:
                continue
            try:
                killer.kill()
            except OSError:
                pass

            i = 0
            while killer.is_alive() and i < 500:
                time.sleep(0.010)
                i += 1
            try:
                killer.force_kill()
            except OSError:
                pass

    def parse_results(self, regex_list: str, output: str, new_kind_results: dict, new_data_results: dict) -> Tuple[
        bool, bool]:
        has_err = False
        has_values = False
        try:
            for result_regex in regex_list:
                for nr in re.finditer(result_regex, output.strip(), re.IGNORECASE):
                    result_type = nr.group("type")

                    kind = nr.group("kind")
                    if kind is None:
                        kind = "time"
                    kind_value = nr.group("kind_value")
                    if result_type is None:
                        result_type = ''
                    n = float(nr.group("value"))
                    mult = nr.group("multiplier")
                    unit = ""
                    if nr.group("unit"):
                        unit = nr.group("unit")
                    if unit.lower() == "sec" or unit.lower() == "s":
                        unit = "s"

                    if unit == "s":
                        if mult == "m":
                            n = n / 1000  # Keep all results in seconds
                        elif mult == "u" or mult == "µ":
                            n = n / 1000000
                        elif mult == "n":
                            n = n / 1000000000
                    else:
                        mult = mult.upper()

                    if mult == "K":
                        n *= 1024
                    elif mult == "M":
                        n *= 1024 * 1024
                    elif mult == "G":
                        n *= 1024 * 1024 * 1024
                    if n != 0 or (self.config.match("accept_zero", result_type)) or kind_value is not None:
                        result_add = self.config.get_bool_or_in("result_add", result_type)
                        result_append = self.config.get_bool_or_in("result_append", result_type)
                        if kind_value:
                            t = float(kind_value)
                            if result_type in new_kind_results.setdefault(kind,{}).setdefault(t, {}):
                                if result_add:
                                    new_kind_results[kind][t][result_type] += n
                                else:
                                    if type(new_kind_results[kind][t][result_type]) is not list:
                                        new_kind_results[kind][t][result_type] = [new_kind_results[kind][t][result_type]]

                                    new_kind_results[kind][t][result_type].append(n)
                            else:
                                new_kind_results[kind][t][result_type] = n
                        else:
                            if result_append:
                                new_data_results.setdefault(result_type,[]).append(n)
                            elif result_type in new_data_results and result_add:
                                new_data_results[result_type] += n
                            else:
                                new_data_results[result_type] = n
                        has_values = True
                    else:
                        print("Result for %s is 0 !" % result_type)
                        has_err = True

        except Exception as e:
            print("Exception while parsing results :")
            has_err = True
            raise e
        return has_err, has_values

    def execute(self, build, run, v, n_runs=1, n_retry=0, allowed_types=SectionScript.ALL_TYPES_SET, do_imports=True,
                test_folder=None, event=None, v_internals={}, before_test = None) \
            -> Tuple[Dict, Dict, str, str, int]:

        # Get address definition for roles from scripts
        self.parse_script_roles()

        # Create temporary folder
        deps_repo = []
        depscripts = [imp.testie.scripts for imp in self.imports]

        allscripts = [item for sublist in [self.scripts] + depscripts for item in sublist]
        for script in allscripts:
            deps_repo.extend(script.get_deps_repos(self.options))
        for repo in deps_repo:
            if repo.version is not None:
                v_internals[repo.reponame.upper() + '_VERSION'] = repo.version
        v.update(v_internals)
        if test_folder is None:
            test_folder = self.make_test_folder()
            f_mine = True
        else:
            f_mine = False
        if not os.path.exists(npf.npf_root() + '/' + test_folder):
            os.mkdir(npf.npf_root() + '/' + test_folder)
        save_path = os.getcwd()
        os.chdir(npf.npf_root() + '/' + test_folder)

        # Build file list
        file_list = []
        if SectionScript.TYPE_INIT in allowed_types:
            initv = {}
            for k, vinit in self.variables.statics().items():
                initv[k] = vinit.makeValues()[0]
            file_list.extend(self.build_file_list(initv, self.role, files=self.init_files))
        else:
            file_list.extend(self.build_file_list(v, self.role))

        n_exec = 0
        n_err = 0

        for imp in self.get_imports():
            imp.testie.parse_script_roles()
            imp.imp_v = {}
            for k, val in imp.testie.variables.statics().items():
                imp.imp_v[k] = val.makeValues()[0]
            imp.imp_v.update(v)
            for late_variables in imp.testie.get_late_variables():
                imp.imp_v.update(late_variables.execute(imp.imp_v, imp.testie))
            if SectionScript.TYPE_INIT in allowed_types:
                file_list.extend(imp.testie.build_file_list(imp.imp_v, imp.get_role(), files=imp.testie.init_files))
            else:
                file_list.extend(imp.testie.build_file_list(imp.imp_v, imp.get_role()))

        self.create_files(file_list, test_folder)

        # Launching the tests in itself
        data_results = OrderedDict()  # dict of result_name -> [val, val, val]
        all_kind_results = {}  # dict of kind -> kind_value -> {result_name -> [val, val, val]}
        m = multiprocessing.Manager()
        all_output = []
        all_err = []
        for i in range(n_runs):
            for i_try in range(n_retry + 1):
                if i_try > 0 and not self.options.quiet:
                    print("Re-try tests %d/%d..." % (i_try, n_retry + 1))
                output = ''
                err = ''

                if before_test:
                    before_test(i,i_try)

                queue = m.Queue()

                event = EventBus(m)

                remote_params = []
                for t, v, role in (
                                          [(imp.testie, imp.imp_v, imp.get_role()) for imp in
                                           self.imports] if do_imports else []) + [
                                      (self, v, None)]:
                  for script in t.scripts:
                    srole = role if role else script.get_role()
                    nodes = npf.nodes_for_role(srole)

                    autokill = m.Value('i', 0) if npf.parseBool(script.params.get("autokill", t.config["autokill"])) else None
                    v["NPF_NODE_MAX"] = len(nodes)
                    for i_node, node in enumerate(nodes):
                      v["NPF_NODE"] = node.get_name()
                      v["NPF_NODE_ID"] = i_node
                      multi = script.multi

                      if multi is None:
                          multi = [0]
                      elif multi == '*':
                          if node.multi:
                              multi = range(1, node.multi + 1)
                          else:
                              multi = [1]
                      elif type(multi) == str:
                          multi = [int(multi)]

                      for i_multi in multi:
                        if autokill is not None:
                            autokill.value = autokill.value + 1
                        if not script.get_type() in (allowed_types.difference(set([SectionScript.TYPE_EXIT]))):
                            continue
                        param = RemoteParameters()
                        param.sudo = script.params.get("sudo", False)

                        v["NPF_MULTI"] = i_multi
                        v["NPF_MULTI_ID"] = i_multi
                        v["NPF_MULTI_MAX"] = node.multi if node.multi is not None else 1
                        v["NPF_ARRAY_ID"] = (i_node * v["NPF_MULTI_MAX"]) + i_multi
                        v["NPF_ARRAY_MAX"] = len(nodes) * v["NPF_MULTI_MAX"]
                        if node.mode == "netns" and i_multi > 0:
                            param.virt = "ip netns exec npfns%d" % i_multi
                            param.sudo = True

                        param.commands = "mkdir -p " + test_folder + " && cd " + test_folder + ";\n" + SectionVariable.replace_variables(
                            v,
                            script.content,
                            self_role = srole, self_node=node,
                            default_role_map= self.config.get_dict(
                                "default_role_map"))
                        param.options = self.options
                        param.queue = queue
                        param.stdin = t.stdin.content
                        timeout = t.config['timeout']
                        if 'timeout' in script.params:
                            timeout = float(script.params['timeout'])
                        if self.config['timeout'] == -1 or self.config['timeout'] > timeout:
                            timeout = self.config['timeout']
                        if timeout == -1 or timeout == "-1":
                            timeout = None

                        param.timeout = timeout
                        script.timeout = timeout
                        param.role = srole
                        param.role_id = i_node
                        param.default_role_map = self.config.get_dict("default_role_map")
                        param.delay = script.delay()
                        deps_bin_path = [repo.get_bin_folder() for repo in script.get_deps_repos(self.options) if
                                         not repo.reponame in self.options.ignore_deps]
                        param.bin_paths = deps_bin_path + [build.get_bin_folder()]
                        param.testdir = test_folder
                        param.event = event
                        param.script = script
                        param.name = script.get_name(True)
                        param.autokill = autokill
                        param.env = OrderedDict()
                        param.env.update(v_internals)
                        param.env.update([(k, v.replace('$NPF_BUILD_PATH', build.repo.get_build_path())) for k, v in
                                          build.repo.env.items()])

                        if self.options.rand_env:
                            param.env['RANDENV'] = ''.join(random.choice(string.ascii_lowercase) for i in range(random.randint(0,self.options.rand_env)))
                        if 'waitfor' in script.params:
                            param.waitfor = script.params['waitfor']

                        remote_params.append(param)

                n = len(remote_params)
                n_exec += n
                if n == 0:
                    break
                try:
                    if self.options.allow_mp:
                        p = multiprocessing.Pool(n)
                        parallel_execs = p.map(_parallel_exec,
                                               remote_params)
                    else:
                        print("Sequential execution...")
                        parallel_execs = []
                        for remoteParam in remote_params:
                            parallel_execs.append(_parallel_exec(remoteParam))

                except KeyboardInterrupt:
                    print("Program is interrupted")
                    if self.options.allow_mp:
                        p.close()
                        p.terminate()

                    if not self.options.preserve_temp:
                        for imp in self.imports:
                            imp.testie.cleanup()
                        self.cleanup()
                    os.chdir('..')
                    if not self.options.preserve_temp:
                        shutil.rmtree(test_folder)
                    else:
                        print("Test files have been preserved in :" + test_folder)
                    sys.exit(1)

                if self.options.allow_mp:
                    p.close()
                    p.terminate()
                worked = False
                critical_failed = False

                for iscript, (r, o, e, c, script) in enumerate(parallel_execs):
                    if r == 0:
                        print("Timeout of %d seconds expired for script %s on %s..." % (
                            script.timeout, script.get_name(), script.get_role()))
                        if not self.options.quiet:
                            if not self.options.show_full:
                                print("stdout:")
                                print(o)
                            print("stderr:")
                            print(e)
                        continue
                    if r == -1:
                        os.chdir('..')
                        if not self.options.preserve_temp and f_mine:
                            shutil.rmtree(test_folder)
                        sys.exit(1)
                    if c != 0:
                        n_err = n_err + 1
                        if npf.parseBool(script.params.get("critical", t.config["critical"])):
                            critical_failed = True
                            print("[ERROR] A critical script failed ! Results will be ignored")
                        print("Bad return code (%d) for script %s on %s ! Something probably went wrong..." % (
                            c, script.get_name(), script.get_role()))
                        if not self.options.quiet:
                            if not self.options.show_full:
                                print("stdout:")
                                print(o)
                            print("stderr:")
                            print(e)
                        continue

                for iparallel, (r, o, e, c, script) in enumerate(parallel_execs):
                    if len(self.scripts) > 1:
                        output += "stdout of script %s on %s :\n" % (script.get_name(), script.get_role())
                        err += "stderr of script %s on %s :\n" % (script.get_name(), script.get_role())

                    if r:
                        worked = True
                        output += o
                        err += e

                if SectionScript.TYPE_EXIT in allowed_types:
                 for s,vlist in [(t.testie,t.imp_v) for t in self.imports] + [(self, v)]:
                  for script in s.get_scripts():
                    if not script.type == SectionScript.TYPE_EXIT:
                        continue
                    exitscripts=script.content
                    role_map = self.config.get_dict("default_role_map")
                    role = None

                    role = script.get_role()

                    for i_node, node in enumerate(npf.nodes_for_role(role, role_map)):
                        vlist["NPF_NODE"] = node.get_name()
                        vlist["NPF_NODE_ID"] = i_node

                        cmd = SectionVariable.replace_variables(
                            vlist,
                            exitscripts,self_role = role, default_role_map = role_map)
                        executor=node.executor
                        ncmd = "mkdir -p " + test_folder + " && cd " + test_folder + ";\n" + cmd
                        try:
                            pid, s_output, s_err, c = executor.exec(cmd=ncmd, options=self.options)
                        except Exception as e:
                            print("An error occured!", e)
                        #print(s_output, s_err)
                        output += s_output
                        err += s_err


                all_output.append(output)
                all_err.append(err)

                if not worked or critical_failed:
                    continue

                if not self.config["result_regex"]:
                    break

                has_values = False
                has_err = False
                new_data_results = {}
                new_kind_results = {}
                new_kind_results.setdefault("time", {})
                regex_list = self.config.get_list("result_regex")

                this_has_err, this_has_value = self.parse_results(regex_list, output, new_kind_results,
                                                                  new_data_results)

                if this_has_err:
                    has_err = True
                if this_has_value:
                    has_values = True
                if hasattr(self, 'pyexit') and allowed_types != set([SectionScript.TYPE_INIT]):
                    vs = {'RESULTS': new_data_results, 'TIME_RESULTS': new_kind_results["time"], 'KIND_RESULTS':new_kind_results}
                    vs.update(v)
                    try:
                        exec(self.pyexit.content, vs)
                    except Exception as e:
                        print("ERROR WHILE EXECUTING PYEXIT SCRIPT:")
                        print(e)


                glob_sync = self.config.get_list("glob_sync")
                glob_min = []
                for g in glob_sync:
                    for kind, kind_results in new_kind_results.items():
                        if kind in glob_sync:
                            mg = min(kind_results.keys())
                            glob_min.append(mg)

                for kind, kind_results in new_kind_results.items():
                  if kind_results:
                    all_kind_results.setdefault(kind,{})
                    if kind in glob_sync:
                        min_kind_value = min(glob_min)
                    else:
                        min_kind_value = min(kind_results.keys())
                    nonzero = set()
                    update = {}
                    all_result_types = set()
                    nz = False
                    accept_zero = not self.config.match("accept_zero", kind)
                    if accept_zero:
                        nz = False

                    last_val = {}
                    acc = self.config.get_list("time_sync")
                    for kind_value, results in sorted(kind_results.items()):
                        if not nz: #We still haven't found a non zero kind_value
                            for result_type, result in results.items():
                                if result_type in self.config.get_list("var_repeat"):
                                    last_val[result_type] = result

                                if result != 0:
                                    nz = True
                                    if (not acc or result_type in acc) and not kind in glob_sync:
                                        min_kind_value = kind_value
                            if not nz:
                                continue
                            else:
                                for result_type, result in last_val.items():
                                    results[result_type] = result

                        for result_type, result in results.items():
                            if result_type in self.config.get_dict("var_n_runs") and i >= int(
                                    self.config.get_dict("var_n_runs")[result_type]):
                                continue
                            nonzero.add(result_type)
                            all_result_types.add(result_type)
                            event_t = Decimal(
                                ("%.0" + str(self.config['time_precision']) + "f") % round(float(kind_value - (min_kind_value if self.config.get_bool_or_in("time_sync", kind) else 0)), int(
                                    self.config['time_precision'])))
                            update.setdefault(event_t, {}).setdefault(result_type, [])
                            update[event_t][result_type].extend(result if type(result) is list else [result])
                            if result_type in self.config.get_list("var_repeat"):
                                # Replicate existing time series for all new incoming time points
                                self.ensure_time(event_t, result_type, all_kind_results[kind])

                    # Replicate new results for every time point
                    for event_t, results in update.items():
                        for result_type, result in results.items():
                            if result_type in self.config.get_list("var_repeat"):
                                self.ensure_time(event_t, result_type, update)

                    for kind_value, results in update.items():
                        for result_type, result in results.items():
                            all_kind_results[kind].setdefault(kind_value, {}).setdefault(result_type, []).extend(result)

                    last_v=0
                    for kind_value, results in update.items():
                        for result_type, result in results.items():
                            if not result:
                                continue
                            last_v = np.mean(result)

                    diff = all_result_types.difference(nonzero)
                    if diff:
                        print("Result for %s is 0 !" % ', '.join(diff))
                        has_err = True
                for result_type, result in new_data_results.items():
                    data_results.setdefault(result_type, []).extend(result if type(result) == list else [result])
                if has_values:
                    break

                for result_type in self.config.get_list('results_expect'):
                    if result_type not in data_results and not result_type in all_kind_results:
                        print("Could not find expected result '%s' !" % result_type)
                        has_err = True

                if len(data_results) + sum([len(r) for k, r in all_kind_results.items()]) == 0:
                    print("Could not find results !")

                    has_err = True
                    continue

                if has_err:
                    if not self.options.show_full:
                        print("stdout:")
                        print(output)
                    print("stderr:")
                    print(err)

        if not self.options.preserve_temp:
            for imp in self.imports:
                imp.testie.cleanup()
            self.cleanup()
        os.chdir(save_path)
        if not self.options.preserve_temp and f_mine:
            shutil.rmtree(test_folder)
        return data_results, all_kind_results, all_output, all_err, n_exec, n_err

    def ensure_time(self, event_t, result_type, update):
        if event_t in update:
            if result_type in update[event_t]:
                return
        else:
            update[event_t] = {}

        # Find the previous point
        prev = None
        mindist = Decimal('Inf')
        for u_t, u_r in update.items():
            if u_t > event_t:
                continue
            if result_type not in u_r:
                continue
            dist = float(event_t) - float(u_t)
            if dist < mindist:
                prev = u_t
                mindist = dist
        if prev is not None:
            update[event_t][result_type] = update[prev][result_type].copy()
        else:
            update[event_t][result_type] = []

    def do_init_all(self, build, options, do_test, allowed_types=SectionScript.ALL_TYPES_SET, test_folder=None,
                    v_internals={}):
        if not build.build(options.force_build, options.no_build, options.quiet_build, options.show_build_cmd):
            raise ScriptInitException()
        if not self.build_deps([build.repo], v_internals=v_internals, no_build=options.no_build):
            raise ScriptInitException()

        if (allowed_types is None or "init" in allowed_types) and options.do_init:
            if not options.quiet:
                print("Executing init scripts...")
            vs = {}
            for k, v in self.variables.statics().items():
                vs[k] = v.makeValues()[0]
            for late_variables in self.get_late_variables():
                vs.update(late_variables.execute(vs, self, fail=False))
            data_results, all_kind_results, output, err, num_exec, num_err = self.execute(build, Run(vs), v=vs, n_runs=1,
                                                                                      n_retry=0,
                                                                                      allowed_types={"init"},
                                                                                      do_imports=True,
                                                                                      test_folder=test_folder,
                                                                                      v_internals=v_internals)

            if num_err > 0:
                if not options.quiet:
                    print("Aborting as init scripts did not run correctly !")
                    print("Stdout:")
                    print("\n".join(output))
                    print("Stderr:")
                    print("\n".join(err))
                raise ScriptInitException()

    def execute_all(self, build, options, prev_results: Dataset = None, do_test=True, on_finish=None,
                    allowed_types=SectionScript.ALL_TYPES_SET, prev_kind_results: Dict[str, Dataset] = None) -> Tuple[
        Dataset, bool]:
        """Execute script for all variables combinations. All tools reliy on this function for execution of the testie
        :param allowed_types:Tyeps of scripts allowed to run. Set with either init, scripts or both
        :param do_test: Actually run the tests
        :param options: NPF options object
        :param build: A build object
        :param prev_results: Previous set of result for the same build to update or retrieve
        :return: Dataset(Dict of variables as key and arrays of results as value)
        """
        if not prev_kind_results:
            prev_kind_results = {}

        init_done = False
        test_folder = self.make_test_folder()

        #All the following paths must be relative to the NPF test folder (that is something like NPF's folder/testie1234567/)
        full_test_folder = npf.from_root(test_folder) + os.sep

        dirname, basename, ext = npf.splitpath(options.output if options.output != 'graph' else options.graph_filename)
        out_path = dirname + os.sep
        v_internals = { 'NPF_REPO':get_valid_filename(build.repo.name),'NPF_ROOT': '../', 'NPF_BUILD': '../' + build.build_path(),
                        'NPF_TESTIE_PATH':os.path.relpath(self.path if self.path else "./",full_test_folder),
                        'NPF_RESULT_PATH':os.path.relpath(build.result_folder(), full_test_folder),
                        'NPF_OUTPUT_PATH':os.path.relpath(out_path, full_test_folder)}

        if not SectionScript.TYPE_SCRIPT in allowed_types:
            # If scripts is not in allowed_types, we have to run the init by force now

            self.do_init_all(build, options, do_test=do_test, allowed_types=allowed_types, v_internals=v_internals)
            if not self.options.preserve_temp:
                shutil.rmtree(test_folder)
            return {}, True

        all_data_results = OrderedDict()
        all_kind_results = OrderedDict()
        # If one first, we first ensure 1 result per variables then n_runs
        if options.onefirst:
            total_runs = [1, self.config["n_runs"]]
        else:
            total_runs = [self.config["n_runs"]]

        for runs_this_pass in total_runs:  # Number of results to ensure for this run
            n = 0
            for root_variables in self.variables.expand(method=options.expand):
                n += 1

                variables= {}
                for imp in self.get_imports():
                  if imp.is_include:
                    for k,v in imp.testie.variables.statics().items():
                        variables[k] = v.makeValues()[0]
                variables.update(root_variables)
                run = Run(variables)
                run.variables.update(build.repo.overriden_variables)
                variables = run.variables.copy()

                for late_variables in self.get_late_variables():
                    variables.update(late_variables.execute({**variables, **v_internals}, testie=self))

                for imp in self.get_imports():
                    for late_variables in imp.testie.get_late_variables():
                        variables.update(late_variables.execute({**variables, **v_internals}, imp.testie))
                r_status, r_out, r_err = self.test_require(variables, build)
                if not r_status:
                    if not self.options.quiet:
                        print("Requirement not met for %s" % run.format_variables(self.config["var_hide"]))
                        if r_out.strip():
                            print(r_out.strip())
                        if r_err.strip():
                            print(r_err.strip())

                    continue

                if prev_results and prev_results is not None and not (options.force_test or options.force_retest):
                    run_results = prev_results.get(run, None)
                    if run_results is None:
                        run_results = {}
                else:
                    run_results = {}

                kind_results = {} #kind->(run_with_time -> results))
                kind_results["time"] = OrderedDict()
                config_time_kinds = self.config.get_list("time_kinds")
                if prev_kind_results and not (options.force_test or options.force_retest):
                    nprev_kind_results = {}
                    for kind, prev_kresults in prev_kind_results.items():
                        if config_time_kinds and not kind in config_time_kinds:
                            continue
                        nprev_kresults = OrderedDict()
                        kind_results.setdefault(kind,OrderedDict())
                        for trun, results in prev_kresults.items():
                            if run.inside(trun):
                                kind_results[kind][trun] = results
                            else:
                                nprev_kresults[trun] = results
                        nprev_kind_results[kind] = nprev_kresults
                    prev_kind_results = nprev_kind_results
                if not run_results and options.use_last and build.repo.url:
                    for version in build.repo.method.get_history(build.version, limit=options.use_last):
                        oldb = Build(build.repo, version)
                        r = oldb.load_results(self)
                        if r and run in r:
                            run_results = r[run]
                            break
                if not kind_results and options.use_last and build.repo.url:
                    for version in build.repo.method.get_history(build.version, limit=options.use_last):
                        oldb = Build(build.repo, version)
                        r = oldb.load_results(self, kind=True)
                        found = False
                        if r:
                            for kind, kr in r.items():
                              kind_results.setdefault(kind, OrderedDict)
                              for r_run, results in kr.items():
                                if run in results:
                                    kind_results[kind][r_run] = results[run]
                                    found = True
                        if found:
                            break

                if run_results:
                    for result_type in self.config.get_list('results_expect'):
                        print("Could not find result :", self.config.get_list('results_expect'))
                        if result_type not in run_results:
                            found = False
                            if prev_kind_results:
                              for kind, kr in prev_kind_results.items():
                                for run_kind, results in kr.items():
                                    if result_type in results:
                                        found = True
                                        continue
                            if not found:
                                print("Missing result type %s, re-doing the run" % result_type)
                                run_results = {}
                                prev_kind_results = {}


                have_new_results = False

                #Compute the minimal number of existing results, so we know how much runs we must do
                l=[]
                dall=True
                n_existing_results=[]
                for result_type, results in run_results.items():
                    if self.config.match("accept_zero", result_type):
                        continue
                    if not results:
                        continue
                    n_existing_results.append(len(results))
                    if len(results) < runs_this_pass:
                        l.append(result_type)
                    else:
                        dall=False

                if len(n_existing_results) == 0:
                    n_existing_results = 0
                else:
                    if options.min_test:
                        n_existing_results = min(n_existing_results)
                    else:
                        n_existing_results = max(n_existing_results)

                n_runs = runs_this_pass - (
                    0 if (options.force_test or options.force_retest) or len(run_results) == 0 else n_existing_results)
                if n_runs > 0 and do_test:
                    if not init_done:
                        self.do_init_all(build, options, do_test, allowed_types=allowed_types, test_folder=test_folder,
                                         v_internals=v_internals)
                        init_done = True

                    def print_header(i, i_try):
                        pass
                    if not self.options.quiet:
                        if len(run_results) > 0:
                            if not dall:
                                print("Results %s are missing some points..." % ", ".join(l))
                        if len(self.variables) > 0:
                            def print_header(i, i_try):
                                n_try=int(self.config["n_retry"])
                                print(run.format_variables(self.config["var_hide"]),
                                  "[%srun %d/%d for test %d/%d]" % (  ("retrying %d/%d " % (i_try + 1,n_try)) if i_try > 0 else "", i+1, n_runs, n, len(self.variables)))
                        else:
                            print("Executing single run...")


                    new_data_results, new_all_kind_results, output, err, n_exec, n_err = self.execute(build, run, variables,
                                                                                                  n_runs,
                                                                                                  n_retry=self.config[
                                                                                                      "n_retry"],
                                                                                                  allowed_types={
                                                                                                      SectionScript.TYPE_SCRIPT, SectionScript.TYPE_EXIT},
                                                                                                  test_folder=test_folder,
                                                                                                  v_internals=v_internals, before_test = print_header)
                    if new_data_results:
                        for result_type, values in new_data_results.items():
                            if values is None:
                                continue
                            if options.force_retest:
                                run_results[result_type] = values
                            else:
                                if result_type in run_results and run_results[result_type] is not None:
                                    run_results[result_type].extend(values)
                                else:
                                    run_results[result_type] = values

                            have_new_results = True
                    if new_all_kind_results:
                        have_new_results = True
                else:
                    if not self.options.quiet:
                        print(run.format_variables(self.config["var_hide"]))

                if len(run_results) > 0:
                    if not self.options.quiet:
                        if len(run_results) == 1:
                            print(list(run_results.values())[0])
                        else:
                            print(", ".join(['{0}: {1}'.format(k, run_results[k]) for k in sorted(run_results)]))

                    all_data_results[run] = run_results
                else:
                    all_data_results[run] = {}

                if have_new_results and sum([len(r) for kind,r in new_all_kind_results.items()]) > 0:
                    for kind, kresults in new_all_kind_results.items():
                      kind_results.setdefault(kind, OrderedDict())
                      for time, results in sorted(kresults.items()):
                        time_run = Run(run.variables.copy())
                        time_run.variables[kind] = time
                        for result_type, result in results.items():
                            rt = kind_results[kind].setdefault(time_run, {}).setdefault(result_type, [])
                            if options.force_retest:
                                rt.clear()
                            rt.extend(result)
                for kind, kresults in kind_results.items():
                  all_kind_results.setdefault(kind, OrderedDict())
                  for result_type, result in kresults.items():
                    all_kind_results[kind][result_type] = result

                if self.options.print_time_results:
                    for kind, kresults in all_kind_results.items():
                        print("%s:" % kind)
                        print(kresults)

                if on_finish and have_new_results:
                    def call_finish():
                        on_finish(all_data_results, all_kind_results)

                    thread = threading.Thread(target=call_finish, args=())
                    thread.daemon = True
                    thread.start()

                # Save results
                if all_data_results and have_new_results:
                    if prev_results or prev_kind_results:
                        if all_data_results[run]:
                            if prev_results is None:
                                prev_results = {}
                            prev_results[run] = all_data_results[run]
                        build.writeversion(self, prev_results, allow_overwrite=True)
                        for kind, kr in kind_results.items():
                            prev_kind_results.setdefault(kind,OrderedDict())
                            prev_kind_results[kind].update(kind_results[kind])
                        build.writeversion(self, prev_kind_results, allow_overwrite=True, kind=True, reload=False)
                    else:
                        build.writeversion(self, all_data_results, allow_overwrite=True)
                        build.writeversion(self, all_kind_results, allow_overwrite=True, kind=True)

        if not self.options.preserve_temp:
            try:
                shutil.rmtree(npf.npf_root() + os.sep + test_folder)
            except PermissionError:
                pass
        else:
            print("Test files have been kept in folder %s" % test_folder)

        return all_data_results, all_kind_results, init_done

    def get_title(self):
        if "title" in self.config and self.config["title"] is not None:
            title = self.config["title"]
        elif hasattr(self, "info"):
            title = self.info.content.strip().split('\n', 1)[0]
        else:
            title = self.filename
        return title

    def reject_outliers(self, data):
        m = self.config["accept_outliers_mult"]
        mean = np.mean(data)
        std = np.std(data)
        data = data[abs(data - mean) <= m * std]
        return data

    @staticmethod
    def expand_folder(testie_path, options, tags=None) -> List['Testie']:
        testies = []
        if not os.path.exists(testie_path):
            print("The testie path %s does not exist" % testie_path)
            return testies
        if os.path.isfile(testie_path):
            testie = Testie(testie_path, options=options, tags=tags)
            testies.append(testie)
        else:
            for root, dirs, files in os.walk(testie_path):
                for filename in files:
                    if filename.endswith(".testie"):
                        try:
                            testie = Testie(os.path.join(root, filename), options=options, tags=tags)
                            testies.append(testie)
                        except Exception as e:
                            print("Error during the parsing of %s :\n%s" % (filename, e))

        filtered_testies = []
        for testie in testies:
            missing_tags = testie.test_tags()
            if len(missing_tags) > 0:
                if not options.quiet:
                    print(
                        "Passing testie %s as it lacks tags %s" % (testie.filename, ','.join(missing_tags)))
            else:
                if testie.test_roles_mapping():
                    filtered_testies.append(testie)
                else:
                    raise Exception(
                        "Roles %s cannot be on the same node ! Please use --cluster argument to set them accross nodes" % ' and '.join(
                            s))

        return filtered_testies

    def test_roles_mapping(self):
        for excludes in self.config.get_list("role_exclude"):
            s = excludes.split('+')
            m = set()
            for role in s:
                nodes = npf.node_for_roles(role)
                for node in nodes:
                  if node in m:
                    return False
                  m.add(node)
        return True

    def parse_script_roles(self):
        """
        Look for parameters of scripts and imports for configuration that affects the test, such as NICs parameters like
        the IP address
        """
        for script in (self.scripts + self.imports):
            for k, val in script.params.items():
                nic_match = re.match(r'(?P<nic_idx>[0-9]+)[:](?P<type>' + NIC.TYPES + '+)', k, re.IGNORECASE)
                if nic_match:
                    npf.nodes_for_role(script.get_role(), self.config.get_dict("default_role_map"))[0].get_nic(
                        int(nic_match.group('nic_idx')))[nic_match.group('type')] = val

    def get_imports(self) -> List[SectionImport]:
        return self.imports

    def get_late_variables(self) -> List[SectionLateVariable]:
        return self.late_variables

    def make_test_folder(self):
        test_folder = "testie%s-%05d" % (datetime.datetime.now().strftime("%y%m%d%H%M"), random.randint(1, 2 << 16))
        os.mkdir( npf.npf_root() + os.sep + test_folder)
        return test_folder



import numpy as np
from typing import Dict, List, Tuple
from collections import OrderedDict

from orderedset._orderedset import OrderedSet
from npf.variable import is_numeric, get_numeric
from npf import npf
import natsort
import csv

class Run:
    def __init__(self, variables):
        self.variables = variables

    def format_variables(self, hide=None):
        if hide is None:
            hide = {}
        s = []
        for k, v in self.variables.items():
            if k in hide: continue
            if type(v) is tuple:
                s.append('%s = %s' % (k, v[1]))
            else:
                s.append('%s = %s' % (k, v))
        return ', '.join(s)

    def print_variable(self, k, default=None):
        v = self.variables.get(k,default)
        if type(v) is tuple:
            return v[1]
        else:
            return v

    def copy(self):
        newrun = Run(self.variables.copy())
        return newrun

    def inside(self, o):
        for k, v in self.variables.items():
            if not k in o.variables:
                return False
            ov = o.variables[k]
            if type(v) is tuple:
                v = v[1]
            if type(ov) is tuple:
                ov = ov[1]
            if is_numeric(v) and is_numeric(ov):
                if not get_numeric(v) == get_numeric(ov):
                    return False
            else:
                if not v == ov:
                    return False
        return True

    def intersect(self, common):
        difs = set.difference(set(self.variables.keys()), common)
        for dif in difs:
            del self.variables[dif]
        return self

    def __eq__(self, o):
        return self.inside(o) and o.inside(self)

    def __hash__(self):
        n = 0
        for k, v in self.variables.items():
            if type(v) is tuple:
                v = v[1]
            if is_numeric(v):
                n += get_numeric(v).__hash__()
            else:
                n += str(v).__hash__()
            n += k.__hash__()
        return n

    def __repr__(self):
        return "Run(" + self.format_variables() + ")"

    def __cmp__(self, o):
        for k, v in self.variables.items():
            if not k in o.variables: return 1
            ov = o.variables[k]
            if is_numeric(v) and is_numeric(ov):
                return get_numeric(v) - get_numeric(ov)

            if type(v) is str or type(ov) is str:
                if str(v) < str(ov):
                    return -1
                if str(v) > str(ov):
                    return 1
            else:
                if v < ov:
                    return -1
                if v > ov:
                    return 1
        return 0

    def __lt__(self, o):
        return self.__cmp__(o) < 0

    def __len__(self):
        return len(self.variables)


Dataset = Dict[Run, Dict[str, List]]
ResultType = str

XYEB = Tuple
AllXYEB = Dict[ResultType, List[XYEB]]

def var_divider(testie: 'Testie', key: str, result_type = None):
    div = testie.config.get_dict_value("var_divider", key, result_type=result_type, default=1)
    if is_numeric(div):
        return float(div)
    if div.lower() == 'g':
        return 1024 * 1024 * 1024
    elif div.lower() == 'm':
        return 1024 * 1024
    elif div.lower() == 'k':
        return 1024
    return 1

def group_val(result, t):
                           if t == 'mean':
                               return np.mean(result)
                           elif t == 'avg':
                               return np.average(result)
                           elif t == 'min':
                               return np.min(result)
                           elif t == 'max':
                               return np.max(result)
                           elif t[:4] == 'perc':
                               return np.percentile(result, int(t[4:]))
                           elif t == 'median' or t == 'med':
                               return np.median(result)
                           elif t == 'std':
                               return np.std(result)
                           elif t == 'nres' or t == 'n':
                               return len(result)
                           elif t == 'first':
                                return result[0]
                           elif t == 'last':
                                return result[-1]
                           elif t == 'all':
                               return result
                           else:
                               print("WARNING : Unknown format %s" % t)
                               return np.nan


def write_output(datasets, statics, options, run_list, kind=None):
    if options.output is None:
        return

    all_result_types = OrderedSet()

    for testie,build,all_results in datasets:
        for run, run_results in all_results.items():
            for result_type,results in run_results.items():
                all_result_types.add(result_type)
    for testie, build, all_results in datasets:
        csvs = OrderedDict()
        for run in run_list:
            results_types = all_results.get(run, OrderedDict())
            for result_type in all_result_types:
                if result_type in csvs:
                    type_filename,csvfile,wr = csvs[result_type]
                else:
                    type_filename = npf.build_filename(testie, build, options.output if options.output != 'graph' else options.graph_filename, statics, 'csv', type_str=result_type, show_serie=(len(datasets) > 1 or options.show_serie), force_ext=True, data_folder=True, prefix = kind + '-' if kind else None)
                    csvfile = open(type_filename, 'w')
                    wr = csv.writer(csvfile, delimiter=' ',
                                quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    csvs[result_type] = (type_filename,csvfile,wr)

                result = results_types.get(result_type,None)

                if result is not None:
                       row = []
                       for t in options.output_columns:
                           if t == 'x':
                               for var,val in run.variables.items():
                                   if var in statics:
                                       continue
                                   row.append(val)
                           elif t == 'all_x':
                               for var,val in run.variables.items():
                                   row.append(val)
                           elif t == 'raw':
                               row.extend(result)
                           else:
                               yval = group_val(result,t)
                               if yval is not None:
                                   try:
                                       it = iter(yval)
                                       row.extend(yval)
                                   except TypeError as te:
                                       row.append(yval)
                       if row:
                           wr.writerow(row)
        for result_type in csvs.keys():
            if options.output is not None:
                print("Output written to %s" % csvs[result_type][0])
                csvs[result_type][1].close()

def convert_to_xyeb(datasets: List[Tuple['Testie', 'Build' , Dataset]], run_list, key, do_x_sort, statics, options, max_series = None, series_sort=None, y_group={}, color=[], kind = None) -> AllXYEB:
    write_output(datasets, statics, options, run_list, kind)
    data_types = OrderedDict()
    all_result_types = OrderedSet()

    for testie,build,all_results in datasets:
        for run, run_results in all_results.items():
            for result_type,results in run_results.items():
                all_result_types.add(result_type)

    for testie, build, all_results in datasets:
        x = OrderedDict()
        y = OrderedDict()
        e = OrderedDict()
        for run in run_list:
            if len(run) == 0:
                xval = build.pretty_name()
            else:
                xval = run.print_variable(key, build.pretty_name())

            results_types = all_results.get(run, OrderedDict())
            for result_type in all_result_types:

                #ydiv = var_divider(testie, "result", result_type) results are now divided before
                xdiv = var_divider(testie, key)
                result = results_types.get(result_type,None)

                if xdiv != 1 and is_numeric(xval):
                    x.setdefault(result_type, []).append(get_numeric(xval) / xdiv)
                else:
                    x.setdefault(result_type, []).append(xval)

                if result is not None:
                    yval = group_val(result, y_group[result_type] if result_type in y_group else ( y_group['result'] if 'result' in y_group else 'mean'))
                    y.setdefault(result_type, []).append(yval)

                    std = np.std(result)
                    mean = np.mean(result)
                    e.setdefault(result_type, []).append((mean, std, result))
                else:
                    y.setdefault(result_type, []).append(np.nan)
                    e.setdefault(result_type, []).append((np.nan, np.nan, [np.nan]))


        for result_type in x.keys():

          try:

            if not do_x_sort:
                ox = x[result_type]
                oy = y[result_type]
                oe = e[result_type]
            else:
                order = np.argsort(x[result_type])
                ox = np.array(x[result_type])[order]
                oy = np.array(y[result_type])[order]
                oe = [e[result_type][i] for i in order]


            data_types.setdefault(result_type, []).append((ox,oy,oe,build))
          except Exception as err:
              print("ERROR while transforming data")
              print(err)
              print("x",x[result_type])
              print("y",y[result_type])
              print("e",e[result_type])

    if series_sort is not None and series_sort != "":
        if type(series_sort) is str and series_sort.startswith('-'):
            inverted = True
            series_sort = series_sort[1:]
        else:
            inverted = False

        new_data_types = OrderedDict()
        for result_type, data in data_types.items():
            avg = []
            max = []
            min = []
            for x,y,e,build in data:
                if not np.isnan(np.sum(y)):
                    avg.append(np.sum(y))
                else:
                    avg.append(0)
                max.append(np.max(y))
                min.append(np.min(y))
            if type(series_sort) is list:
                ok = True
                for i,so in enumerate(series_sort):

                    if is_numeric(so):
                        o = so
                        if o >= len(data):
                            print("ERROR: sorting for %s is invalid, %d is out of range" % (result_type,o))
                            ok = False
                            break
                    elif so in [x for x,y,e,build in data]:
                        o = [x for x,y,e,build in data].index(so)
                    elif so in [build.pretty_name() for x,y,e,build in data]:
                        o = [build.pretty_name() for x,y,e,build in data].index(so)
                    else:
                            print("ERROR: sorting for %s is invalid, %s is not in list" % (result_type,so))
                            ok = False
                            break
                    series_sort[i] = o

                if ok:
                    order = series_sort
                else:
                    order = np.argsort(np.asarray(avg))
            elif series_sort == 'avg':
                order = np.argsort(np.asarray(avg))
            elif series_sort == 'max':
                order = np.argsort(- np.asarray(max))
            elif series_sort == 'min':
                order = np.argsort(np.asarray(min))
            elif series_sort == 'natsort':
                order = natsort.index_natsorted(data,key=lambda x: x[3].pretty_name())
            elif series_sort == 'color':
                order = np.argsort(color)
            else:
                raise Exception("Unknown sorting : %s" % series_sort)

            if inverted:
                order = np.flip(order,0)

            data = [data[i] for i in order]
            new_data_types[result_type] = data
        data_types = new_data_types

    if max_series:
        new_data_types = OrderedDict()
        for i,(result_type,data) in enumerate(data_types.items()):
            new_data_types[result_type] = data[:max_series]
        data_types = new_data_types

    return data_types





import os

import numpy as np
import pydotplus as pydotplus
from orderedset import OrderedSet
from sklearn import tree
from collections import OrderedDict

from typing import List

from npf.build import Build
from npf.testie import Testie
from npf.types.dataset import Dataset
from npf import npf

class Statistics:
    @staticmethod
    def run(build: Build, all_results: Dataset, testie: Testie, max_depth=3, filename=None):
        print("Building dataset...")
        dataset = Statistics.buildDataset(all_results, testie)
        for result_type, X, y, dtype in dataset:
            if len(dataset) > 1:
                print("Statistics for %s" % result_type)
            print("Learning dataset built with %d samples and %d features..." % (X.shape[0], X.shape[1]))
            clf = tree.DecisionTreeRegressor(max_depth=max_depth)
            clf = clf.fit(X, y)

            if max_depth is None or max_depth > 8:
                print("No tree graph when maxdepth is > 8")
            else:
                dot_data = tree.export_graphviz(clf, out_file=None, filled=True, rounded=True, special_characters=True,
                                                feature_names=dtype['names'])
                graph = pydotplus.graph_from_dot_data(dot_data)


                f = npf.build_filename(testie, build, filename if not filename is True else None, {}, 'pdf', result_type, show_serie=False, suffix="clf")
                graph.write(f, format=os.path.splitext(f)[1][1:])
                print("Decision tree visualization written to %s" % f)

            vars_values = OrderedDict()

            print("")
            for i, column in enumerate(X.T):
                varname = dtype['names'][i]
                vars_values[varname] = set([v for v in np.unique(column)])

            print("")
            print("Feature importances :")
            # noinspection PyUnresolvedReferences
            l = list(zip(dtype['names'], clf.feature_importances_))
            l.sort(key=lambda x: x[1])
            for key, f in l:
                if len(vars_values[key]) > 1:
                    print("  %s : %0.4f" % (key, f))

            print('')
            print("Better :")
            best = X[y.argmax()]
            print("  ", end='')
            for i, name in enumerate(dtype['names']):

                print("%s = %s, " % (name, best[i] if (dtype['values'][i] is None) else best[i] if type(best[i]) is np.str_ else dtype['values'][i][int(best[i])]), end='')
            print(' : %.02f' % y.max())

            print('')
            print("Means and std/mean per variables :")
            for i, (k, vals) in enumerate(vars_values.items()):
                if len(vals) is 1:
                    continue
                print("%s :" % k)
                for v in sorted(vals):
                    vs = v if (dtype['values'][i] is None) else dtype['values'][i][int(v)]
                    tot = 0
                    n = 0
                    for ic in range(X.shape[0]):
                        if X[ic,i] == v:
                            tot += y[ic]
                            n += 1
                    if n == 0:
                        print("  %s : None" % vs)
                    else:
                        print("  %s : %.02f, " % (vs, tot / n))
                print("")

    @classmethod
    def buildDataset(cls, all_results: Dataset, testie: Testie) -> List[tuple]:
        dtype = testie.variables.dtype()
        y = OrderedDict()
        dataset = []
        for i, (run, results_types) in enumerate(all_results.items()):
            vars = list(run.variables.values())
            if not results_types is None and len(results_types) > 0:
                dataset.append(vars)
                for result_type, results in results_types.items():
                    r = np.mean(results)
                    y.setdefault(result_type, []).append(r)

        dtype['values'] = [None] * len(dtype['formats'])
        for i, f in enumerate(dtype['formats']):
            if f is str:
                dtype['formats'][i] = int
                values = OrderedSet()
                for row in dataset:
                    values.add(row[i])
                    row[i] = values.index(row[i])
                dtype['values'][i] = list(values)
        X = np.array(dataset, ndmin=2)

        lset = []
        for result_type, v in y.items():
            lset.append((result_type, X, np.array(v),dtype))
        return lset



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires=[
    'require-python-3',
    'pandas',
    'numpy',
    'regex',
    'matplotlib',
    'gitpython',
    'typing',
    'pydotplus',
    'scipy',
    'sklearn',
    'orderedset',
    'paramiko',
    'asteval',
    'cryptography',
    'gitdb',
    'pyasn1',
    'natsort',
    'webcolors',
    'colorama',
    'pygtrie'
        ]

setuptools.setup(
    name="npf",
    version="1.0.2",
    author="Tom Barbette",
    author_email="t.barbette@gmail.com",
    install_requires=install_requires,
    description="NPF",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tbarbette/npf",
    packages=setuptools.find_packages(),
    py_modules=['npf_run','npf_compare','npf_watch'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points = {
              'console_scripts': [
                  'npf-run=npf_run:main',
                  'npf-compare=npf_compare:main',
                  'npf-watch=npf_watch:main',
              ],
          },
)



import multiprocessing
import os,errno
import time
from multiprocessing import Queue
from typing import List
import paramiko
from .executor import Executor
from ..eventbus import EventBus
from paramiko.buffered_pipe import PipeTimeout
import socket

class SSHExecutor(Executor):

    def __init__(self, user, addr, path, port):
        super().__init__()
        self.user = user
        self.addr = addr
        self.path = path
        self.port = port
        self.ssh = False
        #Executor should not make any connection in init as parameters can be overwritten afterward

    def get_connection(self, cache=True):
        import paramiko
        if cache and self.ssh:
            return self.ssh
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.addr, username=self.user, port=self.port)
        if cache:
            self.ssh = ssh
        return ssh


    def exec(self, cmd, bin_paths : List[str] = None, queue: Queue = None, options = None, stdin = None, timeout=None, sudo=False, testdir=None, event=None, title=None, env={}, virt = "", raw = False):
        if not title:
            title = self.addr
        if not event:
            event = EventBus()
        path_list = [p if os.path.isabs(p) else self.path+'/'+p for p in (bin_paths if bin_paths is not None else [])]
        if options and options.show_cmd:
            print("Executing on %s%s (PATH+=%s) :\n%s" % (self.addr,(' with sudo' if sudo and self.user != "root" else ''),':'.join(path_list) + (("NS:"  + virt) if virt else ""), cmd.strip()))

        pre = 'cd '+ self.path + ';'

        if self.path:
            env['NPF_ROOT'] = self.path

        for k,v in env.items():
            if v is not None:
                pre += 'export ' + k + '='+v+'\n'
        if path_list:
            path_cmd = 'export PATH="%s:$PATH"\n' % (':'.join(path_list))
        else:
            path_cmd = ''

        if raw:
            unbuffer = ""
        else:
            unbuffer = "unbuffer"
            if stdin is not None:
                unbuffer = unbuffer + " -p"

        if sudo and self.user != "root":
            cmd = "mkdir -p "+testdir+" && sudo -E " + virt +" "+unbuffer+" bash -c '"+path_cmd + cmd.replace("'", "'\"'\"'") + "'";
        else:
            cmd = virt +" "+unbuffer+" bash -c '"+path_cmd + cmd.replace("'", "'\"'\"'") + "'";
            #pre = path_cmd + pre

        try:
            ssh = self.get_connection(cache=False)

            ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command("echo $$;"+ pre + cmd,timeout=timeout, get_pty=True)
            rpid = int(ssh_stdout.readline())
            if stdin is not None:
                ssh_stdin.write(stdin)
            channels = [ssh_stdout, ssh_stderr]
            output = ['','']

            pid = os.getpid()
            step = 0.2
            for channel in channels:
                channel.channel.setblocking(False)

            while not event.is_terminated() and not ssh_stdout.channel.exit_status_ready():
                try:
                    line = None
                    while ssh_stdout.channel.recv_ready() or ssh_stderr.channel.recv_ready():
                        for ichannel,channel in enumerate(channels[:1]):
                            if channel.channel.recv_ready():
                                try:
                                    line = channel.readline()
                                    if options and options.show_full:
                                        self._print(title, line, False)
                                    self.searchEvent(line, event)
                                    output[ichannel] += line
                                except UnicodeDecodeError:
                                    print("Could not decode SSH input")
                    else:
                        event.wait_for_termination(step)
                        if timeout is not None:
                            timeout -= step
                except PipeTimeout:
                    pass
                except socket.timeout:
                    self._print(title, "Interrupted by timeout", True)
                    event.wait_for_termination(step)
                    if timeout is not None:
                        timeout -= step
                except KeyboardInterrupt:
                    event.terminate()
                    return -1, out, err
                if timeout is not None:
                    if timeout < 0:
                        event.terminate()
                        pid = 0
                        break
            if event.is_terminated():
                if not ssh_stdin.channel.closed:
                    ssh_stdin.channel.send(chr(3))
                    ssh.exec_command("kill "+str(rpid))
#                   unneeded ssh.exec_command("kill $(ps -s  "+str(rpid)+" -o pid=)" )
                    i=0
                    ssh_stdout.channel.status_event.wait(timeout=1)

                ret = 0 #Ignore return code because we kill it before completion.
                ssh.close()
            else:
                ret = ssh_stdout.channel.recv_exit_status()

            for ichannel,channel in enumerate(channels):
                for line in channel.readlines():
                    if options and options.show_full:
                        self._print(title, line, False)
                    self.searchEvent(line, event)
                    output[ichannel] += line

            return pid,output[0], output[1],ret
        except socket.gaierror as e:
            print("Error while connecting to %s" % self.addr)
            print(e)
            return 0,'','',-1
        except paramiko.ssh_exception.SSHException as e:
            print("Error while connecting to %s" % self.addr)
            print(e)
            return 0,'','',-1

    def writeFile(self,filename,path_to_root,content):
        f = open(filename, "w")
        f.write(content)
        f.close()

        try:
            with paramiko.SSHClient() as ssh:
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                ssh.load_system_host_keys()
                try:
                    ssh = self.get_connection()
                except Exception as e:
                    print("Cannot connect to %s with username %s" % (self.addr,self.user))
                    raise e

                transport = ssh.get_transport()
                with transport.open_channel(kind='session') as channel:
                    channel.exec_command('mkdir -p %s/%s' % (self.path, path_to_root))
                    if channel.recv_exit_status() != 0:
                        return False
                with transport.open_channel(kind='session') as channel:
                    channel.exec_command('cat > %s/%s/%s' % (self.path,path_to_root,filename))
                    channel.sendall(content)
                    channel.shutdown_write()
                    return channel.recv_exit_status() == 0
        except paramiko.ssh_exception.SSHException as e:
            print("Error while connecting to %s" % self.addr)
            raise e

    def sendFolder(self, path, local=None):
        try:
            with paramiko.SSHClient() as ssh:
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                ssh.load_system_host_keys()
                try:
                    ssh = self.get_connection()
                except Exception as e:
                    print("Cannot connect to %s with username %s" % (self.addr,self.user))
                    raise e

                transport = ssh.get_transport()

                sftp = paramiko.SFTPClient.from_transport(transport)

                ignored = ['.git', '.vimhistory']
                def _send(path):
                    total = 0
                    rlist = sftp.listdir(self.path + path)
                    lpath = lpath if not local else local + os.sep + path
                    for entry in os.scandir(lpath):
                        if entry.is_file():
                            remote = self.path + path + '/' + entry.name
                            if not entry.name in rlist or entry.stat().st_size != sftp.stat(remote).st_size:
                                try:
                                    es = entry.stat()
                                    sftp.put(lpath + '/' + entry.name, remote)
                                    sftp.chmod(remote, es.st_mode)
                                    total += es.st_size
                                except FileNotFoundError:
                                    raise(Exception("Could not send %s to %s"  % (path + '/' + entry.name, remote)))
                        else:
                            if entry.name in ignored:
                                continue
                            if entry.name not in rlist:
                                sftp.mkdir(self.path + path + '/' + entry.name, mode=0o777 )
                                total += _send(path +'/'+entry.name + '/')
                    return total
                curpath = ''
                total = 0
                for d in path.split('/'):
                    curpath = curpath + d + '/'
                    lcurpath = curpath if not local else local + os.sep + curpath
                    if not os.path.isdir(lcurpath):
                        remote = self.path + path
                        try:
                            sftp.stat(remote)
                        except IOError as e:
                            if e.errno is errno.ENOENT:
                                lpath = path if not local else local + os.sep + path
                                es = os.stat(lpath)
                                sftp.put(lpath, remote)
                                sftp.chmod(remote, es.st_mode)
                                total += es.st_size
                        finally:
                            sftp.close()
                        return total
                    try:
                        f = self.path + '/' + curpath
                        sftp.stat(f)
                    except FileNotFoundError:
                        try:
                            sftp.mkdir(f,mode=0o777)
                        except IOError as e:
                            print("Could not make folder %s" % f)
                            raise e
                    except PermissionError as e:
                        print("Could not make folder %s" % f)
                        raise e


                total += _send(path)

                sftp.close()
                return total

        except paramiko.ssh_exception.SSHException as e:
            print("Error while connecting to %s" % self.addr)
            raise e



from npf.testie import Testie


class Module(Testie):
    def __init__(self, testie_path, options, parent, section, tags=None, role=None):
        if section.is_include:
            testie_path = parent.path +'/'+ testie_path
        super().__init__(testie_path, options, tags, role)

